#include "pqOptUMesh.h"
#include "pqPropertiesPanel.h"
#include "pvOptFuncs.cxx"
#include "pvOptFuncs.h"
#include "stlFuncs.cxx"

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <sstream>

#include <QGridLayout>
#include <QPushButton>
#include <QString>
#include <QStringList>
#include <QObject>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QWidget>
#include <QLabel>
#include <QMessageBox>
#include <QFuture>
#include <QtConcurrent/QtConcurrent>

#include <vtksys/SystemTools.hxx>
      
//-----------------------------------------------------------------------------

pqOptUMesh::pqOptUMesh(
  vtkSMProxy* smproxy, vtkSMProperty* smproperty, QWidget* parentObject)
  : Superclass(smproxy, parentObject)
{
  this->setShowLabel(false);

  QGridLayout* gridLayout = new QGridLayout(this);
  gridLayout->setMargin(pqPropertiesPanel::suggestedMargin());
  gridLayout->setHorizontalSpacing(pqPropertiesPanel::suggestedHorizontalSpacing());
  gridLayout->setVerticalSpacing(pqPropertiesPanel::suggestedVerticalSpacing());
  gridLayout->setColumnStretch(1, 1);
  QLabel* spacer = new QLabel(this);

  //link case file button
  QPushButton* caseFileHandlerButton = new QPushButton(this);
  caseFileHandlerButton->setText(tr("Select Output Folder"));
  connect(caseFileHandlerButton, SIGNAL(clicked()), SLOT(caseFileHandler()));
  gridLayout->addWidget(caseFileHandlerButton);

  //About button
  QPushButton* aboutButton = new QPushButton(this);
  aboutButton->setText(tr("About pvOptUMesh"));
  connect(aboutButton, SIGNAL(clicked()), SLOT(aboutMessage()));
  gridLayout->addWidget(aboutButton);

  //Select tri-surface mesh file
  QPushButton* selectSurfaceButton = new QPushButton(this);
  selectSurfaceButton->setText(tr("Select Tri-Surface Mesh"));
  connect(selectSurfaceButton, SIGNAL(clicked()), SLOT(selectSurface()));
  gridLayout->addWidget(selectSurfaceButton);

  QPushButton* generateMeshButton = new QPushButton(this);
  generateMeshButton->setText(tr("Generate Mesh"));
  connect(generateMeshButton, SIGNAL(clicked()), SLOT(generateMesh()));
  gridLayout->addWidget(generateMeshButton);

  QLabel* blockMeshCellsLabel = new QLabel(tr("Block Mesh Cells"), this);
  gridLayout->addWidget(blockMeshCellsLabel);
  blockMeshCells = new QLineEdit(this);
  blockMeshCells->setText("4000");
  gridLayout->addWidget(blockMeshCells);

  goodDocker = new bool();
  meshGenComplete = new bool();
  *goodDocker = false;
  flowIterationsValue = 1000;
  preProcessFile = "preProcessing_snappyHexMesh.sh";
  meshGenLog = "logMeshGeneration.txt";
  caseFileDir = new QString();
}

pqOptUMesh::~pqOptUMesh()
{
}

//-----------------------------------------------------------------------------

// about pvOptUMesh message
void pqOptUMesh::aboutMessage()
{
  QMessageBox about;
  about.setStyleSheet("QLabel{width: 400px; min-width: 600px; max-width: 400px;}");
  about.setText("<b><p align='center'><big>pvOptUMesh<br></b></p></big>");
  about.setStandardButtons(QMessageBox::Ok);
  about.setInformativeText("<p align='center'>A Paraview Mesh Generation GUI Plugin<br>Version: 0.1.0<br><br>© 2021 iDesign Lab, Iowa State University. All rights reserved.<br><br>Developers: Andrew Thomas and Ping He</p>");
  int ret = about.exec();
}

void pqOptUMesh::caseFileHandler()
{
  paths casePaths = pvOptFuncs::linkCaseFile(this);
  *caseFileDir = casePaths.caseFileDir;
  rootPath = casePaths.rootPath;
  dockerMessage();
}

// check if docker is good
void pqOptUMesh::dockerMessage()
{
  dockerFound info = pvOptFuncs::checkDocker(rootPath);
  *goodDocker = info.goodDocker;

  QMessageBox dockerInfo;
  dockerInfo.setText(info.success ? (*goodDocker ? tr("       Docker Successfully Found       ") : tr("      Docker Not Found     ")) : tr("       Error -- No Docker Info       "));
  dockerInfo.setStandardButtons(QMessageBox::Ok);
  dockerInfo.setInformativeText(*goodDocker ? tr(info.infoStr.c_str()) : tr("Without Docker, executing mesh generation script is impossible"));
  int ret = dockerInfo.exec();
}

void pqOptUMesh::selectSurface()
{
  if (caseFileDir->size()<1)
  {
    caseFileHandler();
  }

  QFileDialog dialog(this, tr("Select Tri-Surface Mesh Files"));
  dialog.setDirectory(rootPath.c_str());
  dialog.setFileMode(QFileDialog::ExistingFiles);

  if (dialog.exec())
  {
	  surfacePaths = dialog.selectedFiles();
  }

  std::string constPath = rootPath + "/constant";
  std::string triSurfacePath = constPath + "/triSurface";

  bool createdConstantDirectory = vtksys::SystemTools::MakeDirectory(triSurfacePath.c_str());
  
  for (int i = 0; i<surfacePaths.size(); i++)
  {
    QFile currentFile = surfacePaths[i];
    QFileInfo fileInfo(currentFile.fileName());
    std::string newPath = triSurfacePath + "/" + fileInfo.fileName().toStdString();
    currentFile.copy(newPath.c_str());
    //STLImport(currentFile.fileName())
    //qDebug() << isBinarySTL();
  }
}

std::string pqOptUMesh::writePreProcessing()
{
  std::ostringstream preProcessPath;
  preProcessPath << rootPath << "/" << preProcessFile;
  #ifdef _WIN32
    std::ofstream preProcessing(preProcessPath.str().c_str(), std::ios::binary);
  #else
    std::ofstream preProcessing(preProcessPath.str().c_str(), std::ios::out);
  #endif
  
  preProcessing<<
  "#!/bin/bash\n"
  "\n"
  "if [ -z \"$WM_PROJECT\" ]; then\n"
  "  echo \"OpenFOAM environment not found, forgot to source the OpenFOAM bashrc?\"\n"
  "  exit\n"
  "fi\n"
  "\n"
  "# pre-processing\n"
  "\n"
  "# generate mesh\n"
  "echo \"Generating mesh..\"\n"
  "\n"
  "echo \"Running snappyHexMesh mesh..\"\n"
  "blockMesh &> "<<meshGenLog<<"\n"
  "surfaceFeatureExtract >> "<<meshGenLog<<"\n"
  "snappyHexMesh -overwrite >> "<<meshGenLog<<"\n"
  "createPatch -overwrite >> "<<meshGenLog<<"\n"
  "renumberMesh -overwrite >> "<<meshGenLog<<"\n"
  "echo \"Generating mesh.. Done\"\n"
  "\n"
  "# copy initial and boundary condition files\n"
  "cp -r 0.orig 0" << std::endl;

  preProcessing.flush();
  preProcessing.close();

  return preProcessPath.str();
}

void pqOptUMesh::controlDict(std::string systemPath)
{
  std::ostringstream controlDictPath;
  controlDictPath << systemPath << "/controlDict";
  std::ofstream controlDictFile(controlDictPath.str().c_str(), std::ios::out);

  controlDictFile << 
  "FoamFile\n"
  "{\n"
  "    version     2.0;\n"
  "    format      ascii;\n"
  "    class       dictionary;\n"
  "    location    \"system\";\n"
  "    object      controlDict;\n"
  "}\n"
  "\n"
  "startFrom       startTime;\n"
  "startTime       0;\n"
  "stopAt          endTime;\n"
  "endTime         "<<flowIterationsValue<<";\n"
  "deltaT          1;\n"
  "writeControl    timeStep;\n"
  "writeInterval   "<<flowIterationsValue<<";\n"
  "purgeWrite      0;\n"
  "writeFormat     ascii;\n"
  "writePrecision  16;\n"
  "writeCompression on;\n"
  "timeFormat      general;\n"
  "timePrecision   16;\n"
  "runTimeModifiable true;\n"
  "DebugSwitches\n"
  "{\n"
  "    SolverPerformance 0;\n"
  "}\n";
  controlDictFile.flush();
  controlDictFile.close();
}

void pqOptUMesh::blockMeshDict(std::string systemPath)
{

  std::ostringstream blockMeshDictPath;
  blockMeshDictPath << systemPath << "/blockMeshDict";
  std::ofstream blockMeshFile(blockMeshDictPath.str().c_str(), std::ios::out);

  blockMeshFile <<
  "FoamFile\n"
  "{\n"
  "    version 2.0;\n"
  "    format ascii;\n"
  "    class dictionary;\n"
  "    location system;\n"
  "    object blockMeshDict;\n"
  "}\n"
  "\n"
  "X1 40;\n"
  "X2 40;\n"
  "X3 20;\n"
  "L 20;\n"
  "M 20;\n"
  "N 10;\n"
  "\n"
  "convertToMeters 1;\n"
  "\n"
  "vertices\n"
  "(\n"
  "    ($X1 $X2  0)\n"
  "    ($X1 $X2  0)\n"
  "    ($X1 $X1  0)\n"
  "    ($X2 $X1  0)\n"
  "    ($X2 $X2  $X1)\n"
  "    ($X1 $X2  $X1)\n"
  "    ($X1 $X1  $X1)\n"
  "    ($X2 $X1  $X1)\n"
  ");\n"
  "\n"
  "blocks  \n"
  "(\n"
  "    hex ( 0 1 2 3 4 5 6 7)\n"
  "    ($L $M $N)\n"
  "    simpleGrading (1 1 1)\n"
  ");\n"
  "\n"
  "edges  \n"
  "(\n"
  ");\n"
  "\n"
  "patches  \n"
  "( \n"
  "  patch farfield_inlet\n"
  "  (\n"
  "      (0 4 7 3)\n"
  "  ) \n"
  "  patch farfield_outlet\n"
  "  (\n"
  "      (1 2 6 5)\n"
  "  ) \n"
  "  patch farfield_bot\n"
  "  (\n"
  "      (0 1 5 4)\n"
  "  ) \n"
  "  patch farfield_top\n"
  "  (\n"
  "      (3 7 6 2)\n"
  "  ) \n"
  "  symmetry sym\n"
  "  (\n"
  "      (0 3 2 1)\n"
  "  ) \n"
  "  patch farfield_front\n"
  "  (\n"
  "      (4 5 6 7)\n"
  "  )\n"
  ");\n"
  "\n"
  "mergePatchPairs  \n"
  "(\n"
  ");" << std::endl;

  blockMeshFile.flush();
  blockMeshFile.close();
}

void pqOptUMesh::snappyHexMeshDict(std::string systemPath)
{
}

void pqOptUMesh::generateMesh()
{
  if (caseFileDir->size()<1)
  {
    caseFileHandler();
  }
  if (surfacePaths.size()<1)
  {
    selectSurface();
  }

  std::string moveCmd;
  bool success = false;
  std::string detailedOutput;
  std::string preProcessingPath;

  pvOptFuncs::executeSystemCmd(moveCmd);

  std::string systemPath = rootPath + "/system";
  bool createdDirectory = vtksys::SystemTools::MakeDirectory(systemPath.c_str());
  controlDict(systemPath);
  blockMeshDict(systemPath);
  //writeSnappyHexMeshDict(systemPath);

  preProcessingPath = writePreProcessing();

  if (*goodDocker)
  {
    std::string cmdPreProcessing = "docker run -i --rm -u dafoamuser --mount \"type=bind,src=" + rootPath +
    ",target=/home/dafoamuser/mount\" -w /home/dafoamuser/mount dafoam/opt-packages:v2.2.7 bash -c " +
    "\". /home/dafoamuser/dafoam/loadDAFoam.sh && ./" + preProcessFile + "\"";

    QFuture<void> futureMeshGen = QtConcurrent::run([=]() {
      pvOptFuncs::executeSystemCmd(cmdPreProcessing);
    });
    futureMeshGen.waitForFinished();

    //Gather output for message box
    std::string logPath = rootPath + "/" + meshGenLog;
    success = true;
    QFile file(logPath.c_str());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
      success = false;
    }
    while (!file.atEnd()) 
    {
      QByteArray line = file.readLine();
      detailedOutput.append(line.toStdString().c_str());
    }
    file.close();
  }

  QMessageBox meshGenInfo;
  meshGenInfo.setText(success ? tr("       Mesh Generation Complete       ") : (*goodDocker ? tr("       Error -- No Output File       ") : tr("Docker Not Installed")));
  meshGenInfo.setStandardButtons(QMessageBox::Ok);
  std::string infoMsg = "";
  if (*goodDocker)
  {
    meshGenInfo.setInformativeText(tr(infoMsg.c_str()));
    meshGenInfo.setDetailedText(tr(detailedOutput.c_str()));
  }
  int ret = meshGenInfo.exec();

  *meshGenComplete = true;
}