#!/usr/bin/env python

import os
import re
import sys
import argparse

import tkinter as Tk
from tkinter import ttk
from tkinter import font as tkFont
from tkinter import filedialog as fd 

import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np

parser = argparse.ArgumentParser(description='PvOptGUI Post Processing Interface')
parser.add_argument('--animated', action='store_true', help='Option to animate the plot for a running optimization')
parser.add_argument('--testing', action='store_true', help='Option to run test, outputs number of iterations and type')
parser.add_argument('-log', type=str, nargs='?', help='The log file to read data from')
parser.add_argument('-clTitle', type=str, nargs='?', help='The title to display during the CL solution animation')
parser.add_argument('-simTitle', type=str, nargs='?', help='The title to display during a single flow solution run')
parser.add_argument('-optTitle', type=str, nargs='?', help='The title to display during the optimization animation')
args = parser.parse_args()

testing = False
optimizationRunning = False

optimizationRunning = args.animated
testing = args.testing
optLog = args.log
alphaTitle = args.clTitle
optTitle = args.optTitle
simTitle = args.simTitle

searchAlphaTitle = alphaTitle
searchOptTitle = optTitle
searchSimTitle = simTitle

printInterval = 10
alpha = 0
numFlow = -1
numAdj = -1
numObj = int()

solveAlphaFound = False
alphaFound = False
optimizationFound = False
solvingFlow = False

flowResLabels = ['Ux', 'Uy', 'Uz', 'p', 'he', 'NuTilda', 'k', 'Omega', 'Epsilon']
objFuncLabels = ['CD', 'CL', 'CM']
colors = ['b','r','g','c','gold','purple','m','y','k','navy','maroon']

def getFloats(line):
    numeric_const_pattern = '[-+]? (?: (?: \d* \. \d+ ) | (?: \d+ \.? ) )(?: [Ee] [+-]? \d+ ) ?'
    rx = re.compile(numeric_const_pattern, re.VERBOSE)
    out = rx.findall(line)
    return out

def listToPlot(vector):
    vec = [[float(i) for i in j] for j in vector]
    return np.array(list(zip(*(vec))))

def getObj(vector):
    n = len(vector)
    vec = np.zeros([numObj,n])
    for i in range(n):
        dat = listToPlot(vector[i])
        l=10
        vec[:,i] = dat[l:l+numObj,-2]
    return vec

def getLog():
    global solveAlphaFound
    global alphaFound
    global optimizationFound
    global optimizationRunning
    global solvingFlow

    global aFlowVectors
    global flowVectors
    global adjointVectors
    global currentFlowVec
    global tempFlowVectors
    global tempAdjointVectors
    global iterationRelation

    global alpha
    global numFlow
    global numAdj
    global numObj
    global flowResShown
    global flowIt
    global adjIt

    searchingAdjoint = False
    searchingFlow = False
    flowIt = -1
    sensIt = -1
    s = 0
    series = -1
    alphaSensIt = 0
    flowLen = 13
    flowResShown = np.zeros(flowLen-4)

    #CL Solution vector
    aFlowVectors = [[[]]] #flow, residuals vs minor iteration
    flowVectors = [[[]]] #flow, residuals vs minor iteration
    adjointVectors = [[[]]] #sens, residual vs minor iteration
    tempFlowVectors = [[]]
    tempAdjointVectors = [[]]
    iterationRelation = [['']] #flow iterations, corresponding adjIt
    r=0
    
    if (os.path.exists(optLog)):
        logFile = open(optLog, 'r')
        lines = logFile.readlines()
        for currentLine in lines:
            if 'Running SolveCL to find alpha that matches target CL' in currentLine:
                alphaFound = True
                alpha = 1
            elif 'Completed! alpha' in currentLine:
                alpha = 0
                flowVectors.pop(0)
                alphaSensIt = len(flowVectors)
                aFlowVectors = flowVectors[:][:][:]
                flowVectors = [[[]]]
                solveAlphaFound = True
            elif 'Optimization Problem -- opt' in currentLine:
                optimizationFound = True
                r+=1
                iterationRelation.append([])
            elif 'Evaluating Objective Functions' in currentLine:
                flowIt = (int(getFloats(currentLine)[0]) - alphaSensIt if solveAlphaFound else int(getFloats(currentLine)[0]))
                searchingFlow = True
                searchingAdjoint = False
                s = -1
                tempFlowVectors = [[0]*flowLen]
                iterationRelation[r].append(flowIt)
            elif 'Evaluating Objective Function Sensitivities' in currentLine:
                adjIt = int(getFloats(currentLine)[0])
                searchingAdjoint = True
                searchingFlow = False
                series = -1   
                iterationRelation[r].append(adjIt)
                r+=1
                iterationRelation.append([])
            elif searchingFlow:
                residualData = getFloats(currentLine)
                if 'U Initial residual:' in currentLine:
                    tempFlowVectors.append([0]*flowLen)
                    s+=1
                    tempFlowVectors[s][0] = (1 if s==0 else s * printInterval)
                    tempFlowVectors[s][1] = residualData[0]
                    tempFlowVectors[s][2] = residualData[1]
                    tempFlowVectors[s][3] = residualData[2]
                    flowResShown[:3] = [1,1,1]
                elif 'p Initial residual:' in currentLine:
                    tempFlowVectors[s][4] = residualData[0]
                    flowResShown[3] = 1 
                elif 'he Initial residual:' in currentLine:
                    tempFlowVectors[s][5] = residualData[0]
                    flowResShown[4] = 1
                elif 'nuTilda Initial residual:' in currentLine:
                    tempFlowVectors[s][6] = residualData[0]
                    flowResShown[5] = 1
                elif 'k Initial residual:' in currentLine:
                    tempFlowVectors[s][7] = residualData[0]
                    flowResShown[6] = 1
                elif 'omega Initial residual:' in currentLine:
                    tempFlowVectors[s][8] = residualData[0]
                    flowResShown[7] = 1
                elif 'epsilon Initial residual:' in currentLine:
                    tempFlowVectors[s][9] = residualData[0]
                    flowResShown[8] = 1
                elif 'CD-part1-force:' in currentLine:
                    tempFlowVectors[s][10] = residualData[1]
                elif 'CL-part1-force:' in currentLine:
                    tempFlowVectors[s][11] = residualData[1]
                elif 'CM-part1-force:' in currentLine:
                    tempFlowVectors[s][12] = residualData[1]
                elif 'Flow Runtime:' in currentLine:
                    flowVectors.append(tempFlowVectors[:][:])

            elif searchingAdjoint:
                adjointData = getFloats(currentLine)
                if 'Main iteration' in currentLine:
                    if 'Main iteration 0' in currentLine: 
                        series += 1; 
                        t=0
                        if (series == 0):
                            tempAdjointVectors = [[0]*2*numObj]
                    if (series == 0):
                        if t != 0 : tempAdjointVectors.append([0]*2*numObj)
                        tempAdjointVectors[t][0] = adjointData[0]
                        tempAdjointVectors[t][1] = adjointData[1]
                    elif (series > 0):
                        if t>=len(tempAdjointVectors): tempAdjointVectors.append([0]*2*numObj)
                        tempAdjointVectors[t][2*series] = adjointData[0]
                        tempAdjointVectors[t][2*series+1] = adjointData[1]
                    t+=1
                elif 'Adjoint Runtime:' in currentLine : 
                    adjointVectors.append(tempAdjointVectors[:][:])

            if 'Solution:' in currentLine:
                if optimizationRunning != 't':
                    optimizationRunning = 0

        logFile.close()
    if flowVectors[0] == [[]] : flowVectors.pop(0)
    if adjointVectors[0] == [[]] : adjointVectors.pop(0)
    currentFlowVec = aFlowVectors if alpha else flowVectors
    for i in range(2) : 
        if len(iterationRelation) > 0 : iterationRelation.pop(-1)

    if not optimizationFound and not alphaFound:
        solvingFlow = True
    else:
        solvingFlow = False

def findLogInfo():
    global numObj
    global searchAlphaTitle
    global searchOptTitle
    global searchSimTitle
    global printInterval

    searchingAdjoint = False
    searchingFlowBC = False
    searchingNormalizedFlow = False
    searchingClAlpha = False
    searchingIC = False
    pIFound = False
    machNum = 0
    reynoldsNum=0
    targetCL = 0.5
    meshCells = 0
    solver = ''
    turbulenceModel = ''
    solveClAlpha = 0

    if (os.path.exists(optLog)):
        logFile = open(optLog, 'r')
        lines = logFile.readlines()
        o=0
        s=0
        t=0
        e=0
        for currentLine in lines:
            readData = np.array([float(i) for i in getFloats(currentLine)])
            if 'Selecting RAS turbulence model' in currentLine and s==0:
                turbulenceModel = currentLine.split()[4]
                s=1
            elif 'Global Cells:' in currentLine:
                meshCells = int(readData[0])
            elif 'solverName' in currentLine:
                solver = currentLine.split()[1].rstrip(';')
            elif 'primalBC' in currentLine:
                searchingFlowBC = True
                flowData = np.zeros([4])
                findFlow = np.zeros([5])
            elif 'normalizeStates' in currentLine:
                searchingFlowBC = False
                searchingNormalizedFlow = True
            elif 'objFunc' in currentLine and o==0:
                searchingNormalizedFlow = False
                searchingAdjoint = True
                machNum = flowData[0] / np.sqrt(flowData[1]*1.4*287.0)
                reynoldsNum = int(flowData[0]*flowData[3]/flowData[2])
            elif 'Evaluating Objective Functions' in currentLine and e==0:
                searchingAdjoint = False
                searchingClAlpha = True
                e=1
            elif 'printInterval' in currentLine and pIFound==False:
                printInterval = int(readData[0])
                pIFound = True
            elif 'Objective Function: calcObjFuncValues' in currentLine:
                searchingIC = True
                t+=1
            elif searchingFlowBC:
                if 'U0' in currentLine:
                    findFlow = np.array([1,0,0])
                elif 'T0' in currentLine:
                    findFlow = np.array([0,1,0])
                elif 'nuTilda0' in currentLine:
                    findFlow = np.array([0,0,1])
                if findFlow[0]==1 and ('value' in currentLine):
                    flowData[0] = readData[1]
                    findFlow = np.array([0,0,0])
                elif findFlow[1]==1 and ('value' in currentLine):
                    flowData[1] = readData[1]
                    findFlow = np.array([0,0,0])
                elif findFlow[2]==1 and ('value' in currentLine):
                    r=10
                    flowData[2] = readData[1]/r
                    findFlow = np.array([0,0,0])
            elif searchingNormalizedFlow:
                if 'p_rgh' in currentLine:
                    flowData[3] = readData[0]
            elif searchingAdjoint:
                if 'part1' in currentLine: 
                    o+=1
                elif 'designVar' in currentLine:
                    numObj = o
            elif searchingClAlpha:
                if 'alpha' in currentLine:
                    solveClAlpha = readData[0]
                    searchingClAlpha = False
            elif searchingIC:
                if 'alpha_0' in currentLine:
                    if t==1 : optimizationAlpha = readData[3]
                    #if t==2 : finalAlpha = readData[3]
                if 'CL' in currentLine:
                    targetCL = readData[1]
                    searchingIC = False
        logFile.close()

        if not solvingFlow:
            searchOptTitle = 'Optimization\nMach Number: ' + str(round(machNum,2)) + ', Reynolds Number: ' + str(reynoldsNum) + ', Target CL: ' + str(targetCL) + '\nMesh Cells: ' + str(meshCells) + ', Solver: ' + str(solver) + ', Turbulence Model: ' + str(turbulenceModel)
            searchAlphaTitle = 'CL Solution\nAlpha0: ' + str(solveClAlpha) + ', Target CL: ' + str(targetCL)
        else:
            searchSimTitle = 'Solving Flow\nAlpha: ' + str(solveClAlpha) + ', Mach Number: ' + str(round(machNum,2)) + ', Reynolds Number: ' + str(reynoldsNum) + '\nMesh Cells: ' + str(meshCells) + ', Solver: ' + str(solver) + ', Turbulence Model: ' + str(turbulenceModel)

def quit():
    plt.close()
    root.quit()
    root.destroy()
    sys.exit()

def change_historyType(event):
    global alpha
    global numFlow
    global numAdj
    numFlow = -1
    numAdj = -1
    histStr = tkHistory.get()
    if histStr == 'CL Solution':
        alpha = 1
    else:
        alpha = 0
    redrawPlot([numFlow, numAdj])

def change_flowIteration(event):
    global numFlow
    global numAdj
    global tkFlowIt
    numFlow = int(-1 if tkFlowIt.get() == '' else tkFlowIt.get())
    if alpha == 0 and isinstance(numFlow,int):
        ex = 0
        for i in iterationRelation[1:]:
            if (numFlow in i[:-1]) and (ex == 0) and isinstance(i[0],int):
                ex = 1
                numAdj = i[-1]
    update_labels()
    redrawPlot([numFlow, numAdj])

def change_adjIteration(event):
    global numAdj
    global numFlow
    global tkAdjIt
    numAdj = int(-1 if tkAdjIt.get() == '' else tkAdjIt.get())
    if alpha == 0 and isinstance(numAdj,int):
        ex = 0
        for i in iterationRelation[1:]:
            if i[-1] == numAdj and ex == 0 and isinstance(i[0],int):
                ex = 1
                numFlow = i[-2]
    update_labels()
    redrawPlot([numFlow, numAdj])

def open_log():
    global optLog
    global alpha
    global alphaFound
    global numFlow
    global numAdj
    global optimizationFound
    global history_box
    global optTitle
    global alphaTitle
    global simTitle
    global optimizationRunning
    global solvingFlow

    optimizationRunning = 0
    alphaFound = False
    optimizationFound = False
    solvingFlow = False
    optLog = fd.askopenfilename()
    getLog()
    findLogInfo()
    optTitle = searchOptTitle
    alphaTitle = searchAlphaTitle
    simTitle = searchSimTitle
    alpha = 1 if alphaFound and not optimizationFound else 0
    numFlow = -1
    numAdj = -1
    update_labels()
    history_box.current(len(history_box['values'])-1)
    redrawPlot([numFlow, numAdj])

def update_labels():
    global adj_iteration
    global flow_iteration
    global currentFlowVec
    global history_box
    global numObj

    if alpha == 1 or len(adj_iteration['values'])==0:
        adj_iteration['values'] = ['']
        adj_iteration.current(0)
    elif alpha == 0 and optimizationFound:
        adj_iteration['values'] = [i for i in range(len(adjointVectors))]
        adj_iteration.current(range(len(adjointVectors))[numAdj])
    currentFlowVec = aFlowVectors if alpha else flowVectors

    if len(flow_iteration['values'])==0:
        flow_iteration['values'] = ['']
        flow_iteration.current(0)
    else:
        flow_iteration['values'] = [i for i in range(len(currentFlowVec))]
        flow_iteration.current(range(len(currentFlowVec))[numFlow])
        
    historyType = []
    if alphaFound: historyType.append('CL Solution')
    if optimizationFound: historyType.append('Optimization')
    if len(history_box['values'])==0: historyType.append('Flow Solution')
    history_box['values'] = historyType

    objVectors = getObj(currentFlowVec)
    if len(objVectors[0,:])>0:
        for i in range(numObj):
            objFuncValue_labels[i].config(text=str('%.8f'%(objVectors[i,numFlow])))
    optObjVectors = getObj(flowVectors)
    if len(optObjVectors[0,:])>0:
        for i in range(numObj):
            baselineObjFuncValue_labels[i].config(text=str('%.8f'%(optObjVectors[i,0])))


def initAnimation():
    global fig
    global figureInfo
    global patches
    global legends
    global flowResShown
    global plotLines

    subfigs = figureInfo[0]
    subFigsNest = figureInfo[1]
    flowAxs = figureInfo[2]
    adjAxs = figureInfo[3]
    itObjAxs = figureInfo[4]
    totObjAxs = figureInfo[5]
    for line in flowAxs.lines : line.remove()
    for line in adjAxs.lines : line.remove()
    for i in itObjAxs : 
        for line in i.lines : line.remove()
    for i in totObjAxs : 
        for line in i.lines : line.remove()

    #Initialize plot lines
    flowLines = [flowAxs.semilogy([],[],label=flowResLabels[i],color=colors[i])[0] for i in range(len(flowResLabels))]
    adjLines = [adjAxs.semilogy([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]
    itObjLines = [itObjAxs[numObj-(i+1)].plot([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]
    totObjLines = [totObjAxs[numObj-(i+1)].plot([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]
    plotLines = np.array([flowLines,adjLines,itObjLines,totObjLines], dtype=list)
    patches = np.sum(plotLines)

    #Initialize Labels
    flowAxs.set_xlabel('Flow Iterations')
    flowAxs.set_ylabel('Flow Residuals')
    flowAxs.grid(True)
    adjAxs.set_xlabel('Adjoint Iterations')
    adjAxs.set_ylabel('Adjoint Residuals')
    adjAxs.grid(True)
    itObjAxs[numObj-1].set_xlabel('Flow Iterations')
    s=numObj-1
    for ax in itObjAxs:
        ax.set_ylabel(objFuncLabels[s])
        ax.grid(True)
        s-=1
    subFigsNest[1].suptitle('Optimization History')
    totObjAxs[numObj-1].set_xlabel('Flow Solution #')
    s=numObj-1
    for ax in totObjAxs:
        ax.set_ylabel(objFuncLabels[s])
        ax.grid(True)
        s-=1

    patches = np.sum(plotLines)
    return patches

def animatePlot(frame):
    global tempFlowVectors
    global tempAdjointVectors
    global adjointVectors
    global currentFlowVec
    global fig
    global patches
    global plotLines
    global ani

    getLog()
    fig.suptitle(alphaTitle if alpha else (simTitle if solvingFlow else optTitle), fontsize=int(root.winfo_height()/ppi+1))
    legends = [[] for i in range(4)]
    flowData = np.array([])

    lns = []
    rolling=[[],[]]
    flowLines = plotLines[0]
    flowAxs.set_title('Flow Solution # ' + str(flowIt), fontsize='large')
    if len(tempFlowVectors) > 1:
        flowData = listToPlot(tempFlowVectors)
        for i in range(len(flowResLabels)):
            if flowResShown[i]==1 and len(flowData[i+1,:])>1:
                r = np.where(flowData[i+1,:]==0.0)[0][0] if np.any(flowData[i+1]==0.0) else None
                flowLines[i].set_data(flowData[0,:r],flowData[i+1,:r])
                lns += [flowLines[i]]
                rolling[0] += [np.min(flowData[i+1,:r])]
                rolling[1] += [np.max(flowData[i+1,:r])]
        legends[0] = flowAxs.legend(lns, [l.get_label() for l in lns],loc='upper right')
        lim=np.max(flowData[0,:])
        flowAxs.set_xlim([-0.05*lim,lim*1.05])
        flowAxs.set_ylim([0.8*np.min(rolling[0]),1.2*np.max(rolling[1])])

    lns = []
    rolling=[[],[]]
    adjLines = plotLines[1]
    adjAxs.set_title('Adjoint Solution # ' + ('' if (alpha or len(adjointVectors) == 0) else str(adjIt)), fontsize='large')
    if len(tempAdjointVectors) > 1 and optimizationFound:
        adjData = listToPlot(tempAdjointVectors)
        for i in range(numObj):
            r = np.where(adjData[2*i+1,:]==0.0)[0][0] if np.any(adjData[2*i+1,1:]==0.0) else -1
            adjLines[i].set_data(adjData[2*i,:r], adjData[2*i+1,:r])
            lns += [adjLines[i]]
            if np.size(adjData[2*i+1,:r]) != 0:
                rolling[0] += [np.min(adjData[2*i+1,:r])]
                rolling[1] += [np.max(adjData[2*i+1,:r])]
        legends[1] = adjAxs.legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')
        lim = max(adjData[0,:])
        adjAxs.set_xlim([-0.05*lim,lim*1.05])
        adjAxs.set_ylim([0.8*np.min(rolling[0]),1.2*np.max(rolling[1])])

    lns = []
    itObjLines = plotLines[2]
    subFigsNest[0].suptitle('Flow Solution # ' + str(flowIt))
    for i in range(numObj): 
        if len(tempFlowVectors) > 1:
            r = np.where(flowData[i+10,:]==0)[0][0] if np.any(flowData[i+10]==0.0) else -1
            itObjLines[i].set_data(flowData[0,:r], flowData[i+10,:r])
            lns += [itObjLines[i]]
            lim=max(flowData[0,:])
            itObjAxs[(numObj-1)-i].set_xlim([-0.05*lim,lim*1.05])
            itObjAxs[(numObj-1)-i].set_ylim([0.97*np.min(flowData[i+10,:r]) if np.min(flowData[i+10,:r]) > 0 else 1.03 * np.min(flowData[i+10,:r]),1.03*np.max(flowData[i+10,:r]) if np.max(flowData[i+10,:r])>0 else 0.97*np.max(flowData[i+10,:r])])
    legends[2] = itObjAxs[0].legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')
    
    lns = []
    totObjLines = plotLines[3]

    if len(flowVectors) > 0:
        objVectors = getObj(flowVectors)
        for i in range(numObj): 
            objMarker = ' ' if len(objVectors[i,:]) > 1 else '.'
            totObjLines[i].set_marker(objMarker)
            totObjLines[i].set_data(range(len(objVectors[i,:])), objVectors[i,:])
            lns += [totObjLines[i]]
            lim = len(objVectors[i])
            totObjAxs[(numObj-1)-i].set_xlim([-0.05*lim,lim*1.05])
            totObjAxs[(numObj-1)-i].set_ylim([0.98*np.min(objVectors[i]) if np.min(objVectors[i])>0 else 1.02*np.min(objVectors[i]),1.02*np.max(objVectors[i]) if np.max(objVectors[i])>0 else 0.98*np.max(objVectors[i])])
        legends[3] = totObjAxs[0].legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')
    
        if len(objVectors[0,:])>0:
            for i in range(numObj):
                objFuncValue_labels[i].config(text=str('%.8f'%(objVectors[i,-1])))
        optObjVectors = getObj(flowVectors)
        if len(optObjVectors[0,:])>0:
            for i in range(numObj):
                baselineObjFuncValue_labels[i].config(text=str('%.8f'%(optObjVectors[i,0])))

    patches = flowLines+adjLines+itObjLines+totObjLines+legends
    if optimizationRunning==0: 
        ani._stop()
        findLogInfo()
        redrawPlot([-1,-1])
    return patches


def redrawPlot(n):
    global fig
    global figureInfo
    global canvas
    global subfigs
    global subFigsNest
    global flowAxs
    global adjAxs
    global itObjAxs
    global totObjAxs
    global currentFlowVec
    global adjointVectors

    currentFlowVec = aFlowVectors if alpha else flowVectors

    fig.suptitle(alphaTitle if alpha else (simTitle if solvingFlow else optTitle), fontsize=int(screen_height/ppi+1))

    #Initialize Labels
    flowAxs.clear()
    flowAxs.set_title('Flow Solution # ' + ('' if len(currentFlowVec) == 0 else str(currentFlowVec.index(currentFlowVec[n[0]]))), fontsize='large')
    flowAxs.set_xlabel('Flow Iterations')
    flowAxs.set_ylabel('Flow Residuals')
    flowAxs.grid(True)
    adjAxs.clear()
    adjAxs.set_title('Adjoint Solution # ' + ('' if (alpha or len(adjointVectors) == 0) else str(adjointVectors.index(adjointVectors[n[1]]))), fontsize='large')
    adjAxs.set_xlabel('Adjoint Iterations')
    adjAxs.set_ylabel('Adjoint Residuals')
    adjAxs.grid(True)
    for i in itObjAxs: i.clear()
    subFigsNest[0].suptitle('Flow Solution # ' + ('' if len(currentFlowVec) == 0  else str(currentFlowVec.index(currentFlowVec[n[0]]))))
    itObjAxs[-1].set_xlabel('Flow Iterations')
    s=numObj-1
    for ax in itObjAxs:
        ax.set_ylabel(objFuncLabels[s])
        ax.grid(True)
        s-=1
    for i in totObjAxs: i.clear()
    subFigsNest[1].suptitle('Optimization History')
    totObjAxs[-1].set_xlabel('Flow Solution #')
    s=numObj-1
    for ax in totObjAxs:
        ax.set_ylabel(objFuncLabels[s])
        ax.grid(True)
        s-=1

    #plot
    lns=[]
    if len(currentFlowVec) > 0:
        if len(currentFlowVec[n[0]]) > 0:
            flowData = listToPlot(currentFlowVec[n[0]])
            for i in range(len(flowResLabels)):
                if np.all(flowData[i+1,1:-1] != 0) :
                    r = np.where(flowData[i+1,:]==0)[0][0]
                    lns+=[flowAxs.semilogy(flowData[0,:r],flowData[i+1,:r],label=flowResLabels[i],color=colors[i])[0]]
            flowAxs.legend(lns, [l.get_label() for l in lns])

    lns=[]
    if len(adjointVectors)>0:
        if len(adjointVectors[numAdj]) > 1 and optimizationFound and not alpha:
            adjData = listToPlot(adjointVectors[numAdj])
            for i in range(numObj):
                r = np.where(adjData[2*i+1,:] == 0)[0][0] if np.any(adjData[2*i+1,:]==0.0) else -1
                lns+=[adjAxs.semilogy(adjData[2*i,:r], adjData[2*i+1,:r],label=objFuncLabels[i],color=colors[i])[0]]
            adjAxs.legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__())

    lns=[[],[]]
    if len(currentFlowVec) > 0:
        objVectors = getObj(currentFlowVec)
        for i in range(numObj): 
            if np.all(flowData[i+10,1:-1] != 0):
                r = np.where(flowData[i+10,:]==0)[0][0]
                lns[0]+=[itObjAxs[(numObj-1)-i].plot(flowData[0,:r], flowData[i+10,:r], label=objFuncLabels[i], color=colors[i])[0]]
            r = n[0]+1 if n[0]!=-1 else None
            objMarker = ' ' if len(objVectors[i,:r]) > 1 else '.'
            lns[1]+=[totObjAxs[(numObj-1)-i].plot(range(len(objVectors[i,:r])), objVectors[i,:r], label=objFuncLabels[i], color=colors[i], marker=objMarker)[0]]
        itObjAxs[-1].legend(lns[0].__reversed__(), [l.get_label() for l in lns[0]].__reversed__())
        totObjAxs[-1].legend(lns[1].__reversed__(), [l.get_label() for l in lns[1]].__reversed__())

    update_labels()
    flow_iteration['values'] = [i for i in range(len(currentFlowVec))]
    adj_iteration['values'] = [i for i in range(len(adjointVectors))]
    canvas.draw()
    canvas.flush_events()


###Run###

while(numObj == 0):
    findLogInfo()
    if numObj==0 and optimizationRunning!=1:
        numObj = 2
getLog()
if not testing:
    matplotlib.use('TkAgg')
    root = Tk.Tk()
    tkHistory = Tk.StringVar(root)
    tkFlowIt = Tk.StringVar(root)
    tkAdjIt = Tk.StringVar(root)

    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    ppi=96
    fig = plt.figure(dpi=ppi, facecolor='white', constrained_layout=True)
    #canvas = Tk.Canvas(master=root, fig)
    canvas = FigureCanvasTkAgg(fig, master=root)
    canvas.draw_idle()
    #Initialize subfigures and plot locations
    subfigs = fig.subfigures(nrows=2,ncols=1,wspace=2.0)
    (flowAxs, adjAxs) = subfigs[0].subplots(1,2)
    #flowAxs.autoscale_on(True)
    subFigsNest = subfigs[1].subfigures(1,2, hspace=None,wspace=None)
    itObjAxs = subFigsNest[0].subplots(numObj, sharex=True)
    totObjAxs = subFigsNest[1].subplots(numObj, sharex=True)
    figureInfo = [subfigs,subFigsNest,flowAxs,adjAxs,itObjAxs,totObjAxs]

    ### tk interface
    root.wm_title('pvOptAirfoil')
    root.protocol('WM_DELETE_WINDOW', quit)
    root.geometry('960x640')
    cols = 4
    rows = 20
    for i in range(rows): root.grid_rowconfigure(i,weight=1)
    for i in range(cols): root.grid_columnconfigure(i,weight=1)
    canvas.get_tk_widget().grid(column=0,row=0,columnspan=cols-1,rowspan=rows-1,sticky='NSEW')
    font = tkFont.Font(family='Helvetica Heavy', size=int(screen_height/ppi))
    toolbar = NavigationToolbar2Tk(canvas, root, pack_toolbar=False)
    toolbar.grid(column=0,row=rows-1)
    toolbar.update()
    main_frame = Tk.Frame(root)
    main_frame.grid(column=cols-1,row=0,rowspan=rows-2,columnspan=1,sticky='NSEW')

    #Simulation
    for i in range(12): main_frame.grid_rowconfigure(i, weight=1)
    for i in range(2): main_frame.grid_columnconfigure(i, weight=1)
    history_label = Tk.Label(main_frame, text='Case History', font=font)
    history_label.grid(column=0, row=1, sticky='NSEW')
    #Simulaton Combo Box
    historyType = []
    if alphaFound: historyType.append('CL Solution')
    if optimizationFound: historyType.append('Optimization')
    if len(historyType)==0: historyType.append('Flow Solution')
    history_box = ttk.Combobox(main_frame, textvariable=tkHistory, values=historyType, font=font)
    if len(historyType)>0 : history_box.current(len(history_box['values'])-1)
    history_box.bind('<<ComboboxSelected>>', change_historyType)
    history_box.grid(column=1, row=1, sticky='nesw')

    #Flow Iteration Title
    flow_iteration_label = Tk.Label(main_frame, text='Flow Solution #', font=font)
    flow_iteration_label.grid(column=0, row=2, sticky='nesw')
    #Flow Iteration Combo Box
    flow_iteration = ttk.Combobox(main_frame, textvariable=tkFlowIt, font=font)
    if len(currentFlowVec)>0: 
        flow_iteration['values'] = [i for i in range(len(currentFlowVec))]
        #flow_iteration.current(currentFlowVec.index(currentFlowVec[numFlow]))
    flow_iteration.bind('<<ComboboxSelected>>',change_flowIteration)
    flow_iteration.grid(column=1, row=2, sticky='nesw')

    #Adjoint Iteration Title
    adj_iteration_label = Tk.Label(main_frame, text='Adjoint Solution #', font=font)
    adj_iteration_label.grid(column=0, row=3, sticky='nesw')
    #Adjoint Iteration Combo Box
    adj_iteration = ttk.Combobox(main_frame, textvariable=tkAdjIt, font=font)
    if len(adjointVectors)>0:
        adj_iteration['values'] = [i for i in range(len(adjointVectors))]
        #adj_iteration.current(len(adj_iteration['values'])-1)
    adj_iteration.bind('<<ComboboxSelected>>',change_adjIteration)
    adj_iteration.grid(column=1, row=3, sticky='nesw')

    #Choose logFile
    file_label = Tk.Label(main_frame, text='Select Different Log', font=font)
    file_label.grid(column=0, row=4, sticky='nesw')
    #file select button
    file_button = Tk.Button(main_frame, text='Open File', command=open_log, font=font)
    file_button.grid(column=1, row=4, sticky='nesw')

    #Latest Obj Func Values
    #frame config
    objFuncRows = 2*numObj+4
    objFunc_frame = Tk.Frame(main_frame, borderwidth = 3, relief='ridge')
    objFunc_frame.grid(column=0,row=6,rowspan=objFuncRows,columnspan=2,sticky='nesw')
    for i in range(objFuncRows): objFunc_frame.grid_rowconfigure(i, weight=1)
    for i in range(2): objFunc_frame.grid_columnconfigure(i, weight=1)
    # values
    baselineObjFunc_label = Tk.Label(objFunc_frame, text='Baseline Function Values', font=font)
    baselineObjFunc_label.grid(column=0, columnspan=2, row=0, sticky='nesw')
    baselineObjFunc_labels = [i for i in range(numObj)]
    baselineObjFuncValue_labels = [i for i in range(numObj)]
    for i in range(numObj):
        baselineObjFunc_labels[i] = Tk.Label(objFunc_frame, text=objFuncLabels[i]+':', font=font)
        baselineObjFunc_labels[i].grid(column=0, row=i+1, sticky='nesw')

        baselineObjFuncValue_labels[i] = Tk.Label(objFunc_frame, text='', font=font, borderwidth = 1, relief='sunken')
        baselineObjFuncValue_labels[i].grid(column=1, row=i+1, sticky='nesw')

    currentObjFunc_label = Tk.Label(objFunc_frame, text='Current Function Values', font=font)
    currentObjFunc_label.grid(column=0, columnspan=2, row=2+numObj, sticky='nesw')
    objFunc_labels = [i for i in range(numObj)]
    objFuncValue_labels = [i for i in range(numObj)]
    for i in range(numObj):
        objFunc_labels[i] = Tk.Label(objFunc_frame, text=objFuncLabels[i]+':', font=font)
        objFunc_labels[i].grid(column=0, row=i+3+numObj, sticky='nesw')

        objFuncValue_labels[i] = Tk.Label(objFunc_frame, text='', font=font, borderwidth = 1, relief='sunken')
        objFuncValue_labels[i].grid(column=1, row=i+3+numObj, sticky='nesw')

    
    #init
    useBlit = False #Blitted will not update titles or axis ticks/labels, only lines and other axes artists can be updated. Blit gives big speed boost
    if optimizationRunning==1:
        ani = FuncAnimation(fig,func=animatePlot,frames=1,init_func=initAnimation,blit=useBlit,repeat_delay=25)
        ani.repeat = True
        ani._start()
    else:
        optTitle = searchOptTitle
        alphaTitle = searchAlphaTitle
        simTitle = searchSimTitle
        root.after(1,redrawPlot([-1,-1]))
    root.mainloop()
else:
    print(str(len(flowVectors))+' flow iterations')
    print(str(len(adjointVectors))+' adj iterations')




